"""Workflow Run is a distributor instance of a declared workflow."""

from __future__ import annotations

import ast
from collections import deque
from datetime import datetime
import json
import logging
import numbers
import time
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    TYPE_CHECKING,
    Union,
)

from jobmon.client.array import Array
from jobmon.client.swarm.swarm_array import SwarmArray
from jobmon.client.swarm.swarm_task import SwarmTask
from jobmon.client.task_resources import TaskResources
from jobmon.core.cluster import Cluster
from jobmon.core.configuration import JobmonConfig
from jobmon.core.constants import TaskStatus, WorkflowRunStatus
from jobmon.core.exceptions import (
    CallableReturnedInvalidObject,
    DistributorNotAlive,
    EmptyWorkflowError,
    TransitionError,
    WorkflowTestError,
)
from jobmon.core.requester import Requester

# avoid circular imports on backrefs
if TYPE_CHECKING:
    from jobmon.client.workflow import Workflow


logger = logging.getLogger(__name__)


class SwarmCommand:
    def __init__(
        self, func: Callable[..., None], *args: List[SwarmTask], **kwargs: Any
    ) -> None:
        """A command to be run by the distributor service.

        Args:
            func: a callable which does work and optionally modifies task instance state
            *args: positional args to be passed into func
            **kwargs: kwargs to be passed into func
        """
        self._func = func
        self._args = args
        self._kwargs = kwargs

    def __call__(self) -> None:
        self._func(*self._args, **self._kwargs)


class WorkflowRun:
    """WorkflowRun enables tracking for multiple runs of a single Workflow.

    A Workflow may be started/paused/ and resumed multiple times. Each start or
    resume represents a new WorkflowRun.

    In order for a Workflow can be deemed to be DONE (successfully), it
    must have 1 or more WorkflowRuns. In the current implementation, a Workflow
    Job may belong to one or more WorkflowRuns, but once the Job reaches a DONE
    state, it will no longer be added to a subsequent WorkflowRun. However,
    this is not enforced via any database constraints.
    """

    def __init__(
        self,
        workflow_run_id: int,
        workflow_run_heartbeat_interval: Optional[int] = None,
        heartbeat_report_by_buffer: Optional[float] = None,
        fail_fast: bool = False,
        wedged_workflow_sync_interval: int = 600,
        fail_after_n_executions: int = 1_000_000_000,
        status: Optional[str] = None,
        requester: Optional[Requester] = None,
    ) -> None:
        """Initialization of the swarm WorkflowRun."""
        self.workflow_run_id = workflow_run_id

        # state tracking
        self._active_states = [
            TaskStatus.QUEUED,
            TaskStatus.INSTANTIATING,
            TaskStatus.LAUNCHED,
            TaskStatus.RUNNING,
        ]
        self.tasks: Dict[int, SwarmTask] = {}
        self.arrays: Dict[int, SwarmArray] = {}
        self.ready_to_run: deque[SwarmTask] = deque()
        self._task_status_map: Dict[str, Set[SwarmTask]] = {
            TaskStatus.REGISTERING: set(),
            TaskStatus.QUEUED: set(),
            TaskStatus.INSTANTIATING: set(),
            TaskStatus.LAUNCHED: set(),
            TaskStatus.RUNNING: set(),
            TaskStatus.DONE: set(),
            TaskStatus.ADJUSTING_RESOURCES: set(),
            TaskStatus.ERROR_FATAL: set(),
        }

        # cache to get same id
        self._task_resources: Dict[int, TaskResources] = {}

        # workflow run attributes
        if status is None:
            status = WorkflowRunStatus.BOUND
        self._status = status
        self._last_heartbeat_time = time.time()

        # flow control
        self.fail_fast = fail_fast
        self.wedged_workflow_sync_interval = wedged_workflow_sync_interval

        # test parameters to force failure
        self._val_fail_after_n_executions = fail_after_n_executions
        self._n_executions = 0

        # optional config
        # config
        config = JobmonConfig()
        if workflow_run_heartbeat_interval is None:
            self._workflow_run_heartbeat_interval = config.get_int(
                "heartbeat", "workflow_run_interval"
            )
        else:
            self._task_instance_heartbeat_interval = workflow_run_heartbeat_interval
        if heartbeat_report_by_buffer is None:
            self._heartbeat_report_by_buffer = config.get_float(
                "heartbeat", "report_by_buffer"
            )
        else:
            self._heartbeat_report_by_buffer = heartbeat_report_by_buffer

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        # This signal is set if the workflow run receives a resume
        self._terminated = False

        self.initialized = False  # Need to call from_workflow or from_workflow_id

    @property
    def status(self) -> Optional[str]:
        """Status of the workflow run."""
        return self._status

    @property
    def done_tasks(self) -> List[SwarmTask]:
        return list(self._task_status_map[TaskStatus.DONE])

    @property
    def failed_tasks(self) -> List[SwarmTask]:
        return list(self._task_status_map[TaskStatus.ERROR_FATAL])

    @property
    def active_tasks(self) -> bool:
        """Based on the task status map, does the workflow run have more work or not.

        If there are no tasks in active states, the fringe is empty, and therefore we should
        error out.
        """
        # To prevent additional compute, return False immediately if set to an error state.
        # Likely done by fail fast or the max execution loops
        if self.status in (WorkflowRunStatus.ERROR, WorkflowRunStatus.TERMINATED):
            return False

        any_active_tasks = any(
            [any(self._task_status_map[s]) for s in self._active_states]
        ) or any(self.ready_to_run)
        return any_active_tasks

    def from_workflow(self, workflow: Workflow) -> None:
        if self.initialized:
            logger.warning("Swarm has already been initialized")
            return

        # construct arrays
        array: Array
        for array in workflow.arrays.values():
            swarm_array = SwarmArray(array.array_id, array.max_concurrently_running)
            self.arrays[array.array_id] = swarm_array

        # construct SwarmTasks from Client Tasks and populate registry
        for task in workflow.tasks.values():
            cluster = workflow.get_cluster_by_name(task.cluster_name)
            fallback_queues = []
            for queue in task.fallback_queues:
                cluster_queue = cluster.get_queue(queue)
                fallback_queues.append(cluster_queue)

            # create swarmtasks
            swarm_task = SwarmTask(
                task_id=task.task_id,
                array_id=task.array.array_id,
                status=task.initial_status,
                max_attempts=task.max_attempts,
                cluster=cluster,
                task_resources=task.original_task_resources,
                compute_resources_callable=task.compute_resources_callable,
                resource_scales=task.resource_scales,
                fallback_queues=fallback_queues,
            )
            self.tasks[task.task_id] = swarm_task

        # create relationships on swarm task
        for task in workflow.tasks.values():
            swarm_task = self.tasks[task.task_id]

            # assign upstream and downstreams
            swarm_task.num_upstreams = len(task.upstream_tasks)
            swarm_task.downstream_swarm_tasks = set(
                [self.tasks[t.task_id] for t in task.downstream_tasks]
            )

            # create array association
            self.arrays[swarm_task.array_id].add_task(swarm_task)

            # assign each task to the correct set
            self._task_status_map[swarm_task.status].add(swarm_task)

            # compute initial fringe
            if swarm_task.status == TaskStatus.DONE:
                # if task is done check if there are downstreams that can run
                for downstream in swarm_task.downstream_swarm_tasks:
                    downstream.num_upstreams_done += 1

        self.last_sync = self._get_current_time()
        self.num_previously_complete = len(self._task_status_map[TaskStatus.DONE])
        self.workflow_id = workflow.workflow_id
        self.max_concurrently_running: int = workflow.max_concurrently_running
        self.dag_id = workflow.dag_id
        self.initialized = True

    def from_workflow_id(self, workflow_id: int, edge_chunk_size: int = 500) -> None:
        if self.initialized:
            logger.warning("Swarm has already been initialized")
            return
        # Log heartbeat prior to starting work
        self._log_heartbeat()
        self.set_workflow_metadata(workflow_id)
        self.set_tasks_from_db(chunk_size=edge_chunk_size)
        self.set_downstreams_from_db(chunk_size=edge_chunk_size)
        # Resumed workflow runs initialized in LINKING state, so update to bound here
        # once tasks are all created and the DAG is in memory
        self._update_status(WorkflowRunStatus.BOUND)
        self.initialized = True

    def set_workflow_metadata(self, workflow_id: int) -> None:
        """Fetch the dag_id and max_concurrently_running parameters of this workflow."""
        _, resp = self.requester.send_request(
            app_route=f"/workflow/{workflow_id}/fetch_workflow_metadata",
            message={},
            request_type="get",
        )

        database_wf = resp["workflow"]
        if not database_wf:
            raise EmptyWorkflowError(f"No workflow found for workflow id {workflow_id}")

        workflow_id, dag_id, max_concurrently_running = database_wf

        self.last_sync = self._get_current_time()
        self.num_previously_complete = len(self._task_status_map[TaskStatus.DONE])
        self.workflow_id = workflow_id
        self.max_concurrently_running = max_concurrently_running
        self.dag_id = dag_id
        logger.info(f"Initialized Swarm(workflow_id={workflow_id}, dag_id={dag_id})")

    def set_tasks_from_db(self, chunk_size: int = 500) -> None:
        """Pull the tasks that need to be run associated with this workflow.

        I.e. all tasks that aren't in DONE state.
        """
        # Fetch metadata for all tasks
        # Keep a cluster registry.
        cluster_registry: Dict[str, Cluster] = {}
        all_tasks_returned = False
        max_task_id = 0
        logger.info("Fetching tasks from the database")
        while not all_tasks_returned:
            # TODO: make this an asynchronous context manager, avoid duplicating code
            if (
                time.time() - self._last_heartbeat_time
            ) > self._workflow_run_heartbeat_interval:
                self._log_heartbeat()
                logger.info(
                    f"Still fetching tasks, {len(self.tasks)} collected so far..."
                )

            _, resp = self.requester.send_request(
                app_route=f"/workflow/get_tasks/{self.workflow_id}",
                message={"max_task_id": max_task_id, "chunk_size": chunk_size},
                request_type="get",
            )
            task_dict = resp["tasks"]
            # Case when no tasks are returned - happen to have 1000 tasks, for example, and
            # 2 chunks of 500. No more work to be done
            if not task_dict:
                return
            elif len(task_dict) < chunk_size:
                # Last chunk has been returned, exit after task creation
                all_tasks_returned = True

            max_task_id = max(task_dict)

            # populate tasks dict, and arrays registry
            for task_id, metadata in task_dict.items():
                (
                    array_id,
                    status,
                    max_attempts,
                    resource_scales,
                    fallback_queues,
                    requested_resources,
                    cluster_name,
                    queue_name,
                    array_concurrency,
                ) = metadata

                # Convert datatypes as appropriate
                task_id = int(task_id)
                resource_scales = ast.literal_eval(resource_scales)
                for resource, scaler in resource_scales.items():
                    if not isinstance(scaler, numbers.Number):
                        if isinstance(scaler, list):
                            resource_scales[resource] = iter(scaler)
                        else:
                            raise ValueError(
                                "Cannot run CLI resume on non-numeric custom resource "
                                f"scales: found {resource} scaler {scaler} in resource "
                                "scales retrieved from the Jobmon DB."
                            )
                fallback_queues = ast.literal_eval(fallback_queues)
                requested_resources = ast.literal_eval(requested_resources)
                # Construct a queue, a cluster, and a task resources object
                try:
                    cluster = cluster_registry[cluster_name]
                except KeyError:
                    cluster = Cluster(
                        cluster_name=cluster_name, requester=self.requester
                    )
                    cluster.bind()
                    cluster_registry[cluster_name] = cluster

                queue = cluster.get_queue(queue_name)
                fallback_queues = [cluster.get_queue(q) for q in fallback_queues]

                task_resources = TaskResources(
                    requested_resources=requested_resources,
                    queue=queue,
                    requester=self.requester,
                )

                st = SwarmTask(
                    task_id=task_id,
                    array_id=array_id,
                    status=status,
                    max_attempts=max_attempts,
                    task_resources=task_resources,
                    cluster=cluster,
                    resource_scales=resource_scales,
                    fallback_queues=fallback_queues,
                )
                self.tasks[task_id] = st

                # Also create arrays
                if array_id not in self.arrays:
                    array = SwarmArray(
                        array_id=array_id, max_concurrently_running=array_concurrency
                    )
                    self.arrays[array_id] = array
                self.arrays[array_id].add_task(st)

                # Add to correct status queue
                self._task_status_map[st.status].add(st)
        logger.info("All tasks fetched")

    def set_downstreams_from_db(self, chunk_size: int = 500) -> None:
        """Pull downstream edges from the database associated with the workflow."""
        # Get edges in a different route to prevent overload

        # Create two maps: one to map node_id -> task_id
        task_node_id_map: Dict[int, int] = {}
        # one to map task_id -> Set(downstream node_ids)
        task_edge_map: Dict[int, Set] = {}

        start_idx, end_idx = 0, chunk_size
        task_ids = list(self.tasks.keys())
        logger.info("Setting dependencies on tasks")
        while start_idx < len(task_ids):
            # Log heartbeats if needed
            if (
                time.time() - self._last_heartbeat_time
            ) > self._workflow_run_heartbeat_interval:
                self._log_heartbeat()
                logger.info("Still fetching edges from the database...")
            task_id_chunk = task_ids[start_idx:end_idx]
            start_idx = end_idx
            end_idx += chunk_size

            _, edge_resp = self.requester.send_request(
                app_route="/task/get_downstream_tasks",
                message={"task_ids": task_id_chunk, "dag_id": self.dag_id},
                request_type="post",
            )
            downstream_tasks = edge_resp["downstream_tasks"]
            # Format is {task_id: (node_id, '[downstream_node_ids]')}
            for task_id, values in downstream_tasks.items():
                node_id, downstream_node_ids = values
                # Convert to Python datatypes
                task_id = int(task_id)
                if downstream_node_ids:
                    downstream_node_ids = json.loads(downstream_node_ids)

                # Assumption: every single node in the downstream edge is not in "D" state
                # Shouldn't be possible to have a downstream node of a task not in "D" state
                # that is complete. If it is, we'll raise unexpected KeyErrors here
                task_node_id_map[node_id] = task_id
                task_edge_map[task_id] = downstream_node_ids

        # Create the dependency graph
        # NOTE: only tasks that are not in status "DONE" are returned. This means the created
        # dependency graph is not the full DAG, only a subset of the tasks that still need to
        # complete.
        logger.info("All edges fetched from the database, starting to build the graph")

        for task_id, swarm_task in self.tasks.items():
            downstream_edges = task_edge_map[task_id]
            if downstream_edges:
                for downstream_node_id in downstream_edges:
                    # Select the appropriate task id
                    downstream_task_id = task_node_id_map[downstream_node_id]
                    # Select the swarm task and add it as a dependency
                    downstream_swarm_task = self.tasks[downstream_task_id]
                    swarm_task.downstream_swarm_tasks.add(downstream_swarm_task)
                    downstream_swarm_task.num_upstreams += 1

        logger.info("Task DAG fully constructed, swarm is ready to run")

    def run(
        self,
        distributor_alive_callable: Callable[..., bool],
        seconds_until_timeout: int = 36000,
        initialize: bool = True,
    ) -> None:
        """Take a concrete DAG and queue al the Tasks that are not DONE.

        Uses forward chaining from initial fringe, hence out-of-date is not
        applied transitively backwards through the graph. It could also use
        backward chaining from an identified goal node, the effect is
        identical.

        Conceptually:
        all tasks in registering state w/ finished upstreams are ready_to_run
        Put tasks in Adjusting state on the ready_to_run queue

        while there are tasks ready_to_run or currently running tasks:
            queue all tasks that are ready_to_run
            wait for some jobs to complete and add downstreams to the ready_to_run queue
            rinse and repeat

        Args:
            distributor_alive_callable: callable that checks whether or not the distributor
                service is still alive.
            seconds_until_timeout: how long to block while waiting for the next task to finish
                before raising an error.
            initialize: whether to initialize (update WorkflowRun to RUNNING and set fringe)

        Return:
            None
        """
        try:
            if initialize:
                logger.info(f"Executing Workflow Run {self.workflow_run_id}")
                self.set_initial_fringe()
                self._update_status(WorkflowRunStatus.RUNNING)

            time_since_last_full_sync = 0.0
            total_elapsed_time = 0.0
            terminating_states = [
                WorkflowRunStatus.COLD_RESUME,
                WorkflowRunStatus.HOT_RESUME,
            ]

            while self.active_tasks:
                # Expire the swarm after the requested number of seconds
                if total_elapsed_time > seconds_until_timeout:
                    raise RuntimeError(
                        f"Not all tasks completed within the given workflow timeout length "
                        f"({seconds_until_timeout} seconds). Submitted tasks will still run, "
                        f"but the workflow will need to be restarted."
                    )

                # check that the distributor is still alive
                if not distributor_alive_callable():
                    raise DistributorNotAlive(
                        "Distributor process unexpectedly stopped. Workflow will error."
                    )

                # If the workflow run status was updated asynchronously, terminate
                # all active task instances and error out.
                if self.status in terminating_states:
                    logger.warning(
                        f"Workflow Run set to {self.status}. Attempting graceful shutdown."
                    )
                    # Active task instances will be set to "K", the processing loop then
                    # keeps running until all of the states are appropriately set.
                    self._terminate_task_instances()

                # if fail fast and any error
                if self.fail_fast and self._task_status_map[TaskStatus.ERROR_FATAL]:
                    logger.info("Failing after first failure, as requested")
                    break

                # fail during test path
                if self._n_executions >= self._val_fail_after_n_executions:
                    raise WorkflowTestError(
                        f"WorkflowRun asked to fail after {self._n_executions} "
                        "executions. Failing now"
                    )

                # process any commands that we can in the time allotted
                loop_start = time.time()
                time_till_next_heartbeat = self._workflow_run_heartbeat_interval - (
                    loop_start - self._last_heartbeat_time
                )
                if self.status == WorkflowRunStatus.RUNNING:
                    self.process_commands(timeout=time_till_next_heartbeat)

                # take a break if needed
                loop_elapsed = time.time() - loop_start
                if loop_elapsed < time_till_next_heartbeat:
                    sleep_time = time_till_next_heartbeat - loop_elapsed
                    time.sleep(sleep_time)
                    loop_elapsed += sleep_time

                # then synchronize state
                if time_since_last_full_sync > self.wedged_workflow_sync_interval:
                    time_since_last_full_sync = 0.0
                    self.synchronize_state(full_sync=True)
                else:
                    time_since_last_full_sync += loop_elapsed
                    self.synchronize_state()

                total_elapsed_time += time.time() - loop_start

        # user interrupt
        except KeyboardInterrupt:
            logger.warning("Keyboard interrupt raised")
            confirm = input("Are you sure you want to exit (y/n): ")
            confirm = confirm.lower().strip()

            if confirm == "y":
                self._update_status(WorkflowRunStatus.STOPPED)
                raise
            else:
                logger.info("Continuing jobmon...")
                seconds_until_timeout = int(seconds_until_timeout - loop_elapsed)
                self.run(
                    distributor_alive_callable, seconds_until_timeout, initialize=False
                )

        # unexpected errors. raise
        except Exception as e:
            try:
                self._update_status(WorkflowRunStatus.ERROR)
            except TransitionError as trans:
                logger.warning(trans)
            raise e

        # no more active tasks
        else:
            # check if done
            if len(self.tasks) == len(self._task_status_map[TaskStatus.DONE]):
                logger.info("All tasks are done")
                self._update_status(WorkflowRunStatus.DONE)

            else:
                if self.status in terminating_states:
                    self._update_status(WorkflowRunStatus.TERMINATED)
                else:
                    self._update_status(WorkflowRunStatus.ERROR)

    def set_initial_fringe(self) -> None:
        """Set initial fringe."""
        for t in [t for t in self._task_status_map[TaskStatus.ADJUSTING_RESOURCES]]:
            self._set_adjusted_task_resources(t)
            self.ready_to_run.append(t)

        for t in [
            task
            for task in self._task_status_map[TaskStatus.REGISTERING]
            if task.all_upstreams_done
        ]:
            self._set_validated_task_resources(t)
            self.ready_to_run.append(t)

    def get_swarm_commands(self) -> Generator[SwarmCommand, None, None]:
        """Generator to get next chunk of work to be done. Must be idempotent."""
        # compute capacities. max - active
        active_tasks: Set[SwarmTask] = set()
        for task_status in self._active_states:
            active_tasks = active_tasks.union(self._task_status_map[task_status])
        workflow_capacity = self.max_concurrently_running - len(active_tasks)
        array_capacity_lookup: Dict[int, int] = {
            aid: array.max_concurrently_running
            - len(active_tasks.intersection(array.tasks))
            for aid, array in self.arrays.items()
        }

        try:
            unscheduled_tasks: List[SwarmTask] = []
            while self.ready_to_run and workflow_capacity > 0:
                # pop the next task off of the queue
                next_task = self.ready_to_run.popleft()
                array_id = next_task.array_id
                task_resources = next_task.current_task_resources

                # check capacity. add to current batch if room.
                current_batch: List[SwarmTask] = []
                array_capacity = array_capacity_lookup[array_id]
                if array_capacity > 0:
                    current_batch.append(next_task)
                    current_batch_size = 1
                    workflow_capacity -= 1
                    array_capacity -= 1

                    # we started a batch. let's try and add compatible tasks
                    for _ in range(len(self.ready_to_run)):
                        # Remove task from the front of the queue.
                        # If compatible, expire from the ready to run queue
                        task = self.ready_to_run.popleft()
                        # check for batch compatible tasks
                        if (
                            workflow_capacity > 0
                            and array_capacity > 0
                            and task.array_id == array_id
                            and task.current_task_resources == task_resources
                            and current_batch_size < 500
                        ):
                            current_batch.append(task)
                            current_batch_size += 1
                            workflow_capacity -= 1
                            array_capacity -= 1
                        else:
                            # Put it back in the queue
                            self.ready_to_run.append(task)

                    # set final array capacity
                    array_capacity_lookup[array_id] = array_capacity

                    yield SwarmCommand(self.queue_task_batch, current_batch)

                # no room. keep track for next time method is called
                else:
                    unscheduled_tasks.append(next_task)

        # make sure to put unscheduled back on queue, even when the generator is closed
        finally:
            self.ready_to_run.extendleft(unscheduled_tasks)

    def process_commands(self, timeout: Union[int, float] = -1) -> None:
        """Processes swarm commands until all work is done or timeout is reached.

        Args:
            timeout: time until we stop processing. -1 means process till no more work
        """
        swarm_commands = self.get_swarm_commands()

        # this way we always process at least 1 command
        loop_start = time.time()
        keep_processing = True
        while keep_processing:
            # run commands
            try:
                # use an iterator so we don't waste compute
                swarm_command = next(swarm_commands)
                swarm_command()

                # if we need a status sync close the generator. next will raise StopIteration
                if not ((time.time() - loop_start) < timeout or timeout == -1):
                    swarm_commands.close()

            except StopIteration:
                # stop processing commands if we are out of commands
                keep_processing = False

    def synchronize_state(self, full_sync: bool = False) -> None:
        self._set_status_for_triaging()
        self._log_heartbeat()
        self._task_status_updates(full_sync=full_sync)
        self._synchronize_max_concurrently_running()

    def _refresh_task_status_map(self, updated_tasks: Set[SwarmTask]) -> None:
        # remove these tasks from old mapping
        for status in self._task_status_map.keys():
            self._task_status_map[status] = (
                self._task_status_map[status] - updated_tasks
            )

        num_newly_completed = 0
        num_newly_failed = 0
        for task in updated_tasks:
            # assign each task to the correct set
            self._task_status_map[task.status].add(task)

            if task.status == TaskStatus.DONE:
                num_newly_completed += 1
                self._n_executions += 1  # a test param

                # if task is done check if there are downstreams that can run
                for downstream in task.downstream_swarm_tasks:
                    downstream.num_upstreams_done += 1
                    if downstream.all_upstreams_done:
                        self._set_validated_task_resources(downstream)
                        self.ready_to_run.append(downstream)

            elif task.status == TaskStatus.ERROR_FATAL:
                num_newly_failed += 1

            elif task.status == TaskStatus.REGISTERING and task.all_upstreams_done:
                self._set_validated_task_resources(task)
                self.ready_to_run.append(task)

            elif task.status == TaskStatus.ADJUSTING_RESOURCES:
                self._set_adjusted_task_resources(task)

                # put at front of queue since we already tried it once
                self.ready_to_run.appendleft(task)

            else:
                logger.debug(
                    f"Got status update {task.status} for task_id: {task.task_id}."
                    "No actions necessary."
                )

        # if newly done report percent done and check if all done
        if num_newly_completed > 0:
            percent_done = round(
                (len(self._task_status_map[TaskStatus.DONE]) / len(self.tasks)) * 100, 2
            )
            logger.info(
                f"{num_newly_completed} newly completed tasks. {percent_done} percent done."
            )

        # if newly failed, report failures and check if we should error out
        if num_newly_failed > 0:
            logger.warning(f"{num_newly_failed} newly failed tasks.")

    def _set_status_for_triaging(self) -> None:
        app_route = f"/workflow_run/{self.workflow_run_id}/set_status_for_triaging"
        self.requester.send_request(
            app_route=app_route, message={}, request_type="post"
        )

    def _log_heartbeat(self) -> None:
        next_report_increment = (
            self._workflow_run_heartbeat_interval * self._heartbeat_report_by_buffer
        )
        app_route = f"/workflow_run/{self.workflow_run_id}/log_heartbeat"
        _, response = self.requester.send_request(
            app_route=app_route,
            message={
                "status": self._status,
                "next_report_increment": next_report_increment,
            },
            request_type="post",
        )
        self._status = response["status"]
        self._last_heartbeat_time = time.time()

    def _update_status(self, status: str) -> None:
        """Update the status of the workflow_run with whatever status is passed."""
        app_route = f"/workflow_run/{self.workflow_run_id}/update_status"
        _, response = self.requester.send_request(
            app_route=app_route,
            message={"status": status},
            request_type="put",
        )
        self._status = response["status"]
        if self.status != status:
            raise TransitionError(
                f"Cannot transition WFR {self.workflow_run_id} from current status "
                f"{self._status} to {status}."
            )

    def _terminate_task_instances(self) -> None:
        """Terminate the workflow run."""
        app_route = f"/workflow_run/{self.workflow_run_id}/terminate_task_instances"
        self.requester.send_request(app_route=app_route, message={}, request_type="put")
        self._terminated = True

    def _set_fail_after_n_executions(self, n: int) -> None:
        """For use during testing.

        Force the TaskDag to 'fall over' after n executions, so that the resume case can be
        tested.

        In every non-test case, self.fail_after_n_executions will be None, and
        so the 'fall over' will not be triggered in production.
        """
        self._val_fail_after_n_executions = n

    def _get_current_time(self) -> datetime:
        app_route = "/time"
        _, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )
        return response["time"]

    def _task_status_updates(self, full_sync: bool = False) -> None:
        """Update internal state of tasks to match the database.

        If no tasks are specified, get all tasks.
        """
        if full_sync:
            message = {}
        else:
            message = {"last_sync": str(self.last_sync)}

        app_route = f"/workflow/{self.workflow_id}/task_status_updates"
        _, response = self.requester.send_request(
            app_route=app_route,
            message=message,
            request_type="post",
        )
        self.last_sync = response["time"]

        new_status_tasks: Set[SwarmTask] = set()
        for current_status, task_ids in response["tasks_by_status"].items():
            task_ids = set(task_ids).intersection(self.tasks)
            for task_id in task_ids:
                task = self.tasks[task_id]
                if current_status != task.status:
                    task.status = current_status
                    new_status_tasks.add(task)
        self._refresh_task_status_map(new_status_tasks)

    def _synchronize_max_concurrently_running(self) -> None:
        app_route = f"/workflow/{self.workflow_id}/get_max_concurrently_running"
        _, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )
        self.max_concurrently_running = response["max_concurrently_running"]

    def queue_task_batch(self, tasks: List[SwarmTask]) -> None:
        first_task = tasks[0]
        task_resources = first_task.current_task_resources
        if not task_resources.is_bound:
            task_resources.bind()

        app_route = f"/array/{first_task.array_id}/queue_task_batch"
        _, response = self.requester.send_request(
            app_route=app_route,
            message={
                "task_ids": [task.task_id for task in tasks],
                "task_resources_id": task_resources.id,
                "workflow_run_id": self.workflow_run_id,
                "cluster_id": first_task.cluster.id,
            },
            request_type="post",
        )
        updated_tasks = set()
        for status, task_ids in response["tasks_by_status"].items():
            for task_id in task_ids:
                task = self.tasks[task_id]
                task.status = status
                updated_tasks.add(task)
        self._refresh_task_status_map(updated_tasks)

    def _set_validated_task_resources(self, task: SwarmTask) -> None:
        # original resources
        task_resources = task.current_task_resources

        # update with dynamic params
        if task.compute_resources_callable is not None:
            dynamic_compute_resources = task.compute_resources_callable()
            if not isinstance(dynamic_compute_resources, dict):
                raise CallableReturnedInvalidObject(
                    f"compute_resources_callable={task.compute_resources_callable} for "
                    f"task_id={task.task_id} returned an invalid type. Must return dict. got "
                    f"{type(dynamic_compute_resources)}."
                )
            requested_resources = task.current_task_resources.requested_resources.copy()
            requested_resources.update(dynamic_compute_resources)
            task.compute_resources_callable = None
            task_resources = TaskResources(
                requested_resources, task.current_task_resources.queue
            )

        # now check if we need to coerce them to valid values
        validated_task_resources = task_resources.coerce_resources()

        # check for a cached version
        validated_resource_hash = hash(validated_task_resources)
        try:
            validated_task_resources = self._task_resources[validated_resource_hash]
        except KeyError:
            self._task_resources[validated_resource_hash] = validated_task_resources

        # now set as current
        task.current_task_resources = validated_task_resources

    def _set_adjusted_task_resources(self, task: SwarmTask) -> None:
        """Adjust the swarm task's parameters.

        Use the cluster API to generate the new resources, then bind to input swarmtask.
        """
        # current resources
        task_resources = task.current_task_resources.adjust_resources(
            resource_scales=task.resource_scales,
            fallback_queues=task.fallback_queues,
        )

        resource_hash = hash(task_resources)
        try:
            task_resources = self._task_resources[resource_hash]
        except KeyError:
            self._task_resources[resource_hash] = task_resources
        task.current_task_resources = task_resources



"""A logical instance of a project or model that will be run many times over."""

from __future__ import annotations

from http import HTTPStatus as StatusCodes
import logging
from typing import Any, Dict, Optional, Tuple, Type, TYPE_CHECKING

from jobmon.client.task_template import TaskTemplate
from jobmon.core.exceptions import InvalidResponse
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeClientToolVersion

if TYPE_CHECKING:
    from jobmon.client.tool import Tool


logger = logging.getLogger(__name__)


class ToolVersion:
    """Represents a logical instance of a project or model that will be run many times over."""

    def __init__(
        self, tool_version_id: int, requester: Optional[Requester] = None
    ) -> None:
        """Instantiate a tool version.

        Args:
            tool_version_id: an integer id associated with a Tool
            requester: communicate with the flask services.
        """
        self.id = tool_version_id
        self._tool: Tool

        self.task_templates: Dict[str, TaskTemplate] = {}

        self.default_compute_resources_set: Dict[str, Dict[str, Any]] = {}
        self.default_resource_scales_set: Dict[str, Dict[str, float]] = {}
        self.default_cluster_name: str = ""
        self.default_max_attempt: Optional[int] = None

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @classmethod
    def get_tool_version(cls: Type[ToolVersion], tool: Tool) -> ToolVersion:
        """Get a new instance of a ToolVersion from the database.

        Args:
            tool: a Tool to get a version from.
            tool_version_id: tool_version_id to get from the database.
        """
        message = {"tool_id": tool.id}
        app_route = "/tool_version"
        return_code, response = tool.requester.send_request(
            app_route=app_route, message=message, request_type="post"
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )
        tool_version = cls.from_wire(response["tool_version"], tool)
        return tool_version

    @classmethod
    def from_wire(cls: Type[ToolVersion], wire_tuple: Tuple, tool: Tool) -> ToolVersion:
        """Convert from the wire format of ToolVersion to an instance.

        Args:
            wire_tuple: Wire format for ToolVersion defined in jobmon.serializers.
            tool: The Tool object to verify the right tool_version based on tool_id.
        """
        tool_version_kwargs = SerializeClientToolVersion.kwargs_from_wire(wire_tuple)

        if tool_version_kwargs["tool_id"] != tool.id:
            raise ValueError(
                "tool_id in wire_tuple does not match tool object. "
                f"Expected {tool.id} in wire_tuple. Got "
                f"{tool_version_kwargs['tool_id']}"
            )

        tool_version = cls(tool_version_kwargs["id"], requester=tool.requester)
        tool_version._tool = tool
        return tool_version

    @property
    def tool(self) -> Tool:
        """The Tool this ToolVersion is associated with."""
        return self._tool

    def load_task_templates(self) -> None:
        """Get all task_templates associated with this tool version from the database."""
        app_route = f"/tool_version/{self.id}/task_templates"
        return_code, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )

        if return_code != StatusCodes.OK:
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST request through route "
                f"{app_route}. Expected code 200. Response content: {response}"
            )
        task_templates = [
            TaskTemplate.from_wire(wire_tuple, self)
            for wire_tuple in response["task_templates"]
        ]
        for task_template in task_templates:
            self.task_templates[task_template.template_name] = task_template
            task_template.load_task_template_versions()

    def get_task_template(self, template_name: str) -> TaskTemplate:
        """Get a single task_template associated with this tool version from the database."""
        task_template = self.task_templates.get(template_name)
        if task_template is None:
            task_template = TaskTemplate.get_task_template(self, template_name)
            task_template.load_task_template_versions()
            self.task_templates[template_name] = task_template
        return task_template

    def update_default_compute_resources(
        self, cluster_name: str, **kwargs: Any
    ) -> None:
        """Update default compute resources in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        compute_resources = {cluster_name: kwargs}
        self.default_compute_resources_set.update(compute_resources)

    def update_default_resource_scales(self, cluster_name: str, **kwargs: Any) -> None:
        """Update default resource scales in place only overridding specified keys.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to modify default values for.
            **kwargs: any key/value pair you want to update specified as an argument.
        """
        resource_scales = {cluster_name: kwargs}
        self.default_resource_scales_set.update(resource_scales)

    def set_default_compute_resources_from_dict(
        self, cluster_name: str, compute_resources: Dict[str, Any]
    ) -> None:
        """Set default compute resources for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            compute_resources: dictionary of default compute resources to run tasks
                with. Can be overridden at task template or task level.
                dict of {resource_name: resource_value}
        """
        self.default_compute_resources_set[cluster_name] = compute_resources

    def set_default_resource_scales_from_dict(
        self, cluster_name: str, resource_scales: Dict[str, float]
    ) -> None:
        """Set default resource scales for a given cluster_name.

        If no default cluster is specified when this method is called, cluster_name will
        become the default cluster.

        Args:
            cluster_name: name of cluster to set default values for.
            resource_scales: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task template or task level.
                dict of {resource_name: scale_value}
        """
        self.default_resource_scales_set[cluster_name] = resource_scales

    def set_default_max_attempts(self, value: int) -> None:
        """Set default max attempts at tool leve.

        Args:
            value: the default max attempts value.
        """
        if value:
            self.default_max_attempt = value
        else:
            logger.info("The default_max_attempt for tool_version can not be None.")

    def __repr__(self) -> str:
        """A representation string for a ToolVersion instance."""
        return (
            f"ToolVersion(tool_version_id={self.id},"
            f"task_templates: {[t for t in self.task_templates.keys()]})"
        )



"""The workflow run is an instance of a workflow."""

from __future__ import annotations

import getpass
import logging
import time
from typing import Optional

from jobmon.client import __version__
from jobmon.core.configuration import JobmonConfig
from jobmon.core.exceptions import WorkflowNotResumable
from jobmon.core.requester import Requester


logger = logging.getLogger(__name__)


class WorkflowRunFactory:
    """A utility class responsible for instantiating workflow run objects.

    This class sends the appropriate resume signals so that the parent workflow
    object is in a state where the newly created workflowrun is ready to run, either on
    resume or not.
    """

    # TODO: workflow run factory is for now mostly a placeholder for various functions
    # useful for instantiating workflow runs either from a workflow or from the CLI.
    # Might want to consider unifying the resume API more, think about how to handle
    # task resource caching and creation as well in order to resume.

    def __init__(self, workflow_id: int, requester: Optional[Requester] = None) -> None:
        """Initialization of client WorkflowRun."""
        self.workflow_id = workflow_id
        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester
        self.workflow_is_resumable = False

    def set_workflow_resume(
        self, reset_running_jobs: bool = True, resume_timeout: int = 300
    ) -> None:
        """Set statuses of the given workflow ID's workflow to a resumable state.

        Move active workflow runs to hot/cold resume states, depending on reset_running_jobs.
        """
        if self.workflow_is_resumable:
            return
        app_route = f"/workflow/{self.workflow_id}/set_resume"
        self.requester.send_request(
            app_route=app_route,
            message={"reset_running_jobs": reset_running_jobs},
            request_type="post",
        )
        # Wait for the workflow to become resumable
        self.wait_for_workflow_resume(resume_timeout)

    def wait_for_workflow_resume(self, resume_timeout: int = 300) -> None:
        # previous workflow exists but is resumable. we will wait till it terminates

        wait_start = time.time()
        while not self.workflow_is_resumable:
            logger.info(
                f"Waiting for resume. "
                f"Timeout in {round(resume_timeout - (time.time() - wait_start), 1)}"
            )
            app_route = f"/workflow/{self.workflow_id}/is_resumable"
            return_code, response = self.requester.send_request(
                app_route=app_route, message={}, request_type="get"
            )

            self.workflow_is_resumable = bool(response.get("workflow_is_resumable"))
            if (time.time() - wait_start) > resume_timeout:
                raise WorkflowNotResumable(
                    "workflow_run timed out waiting for previous "
                    "workflow_run to exit. Try again in a few minutes."
                )
            else:
                sleep_time = round(float(resume_timeout) / 10.0, 1)
                time.sleep(sleep_time)

    def reset_task_statuses(self, reset_if_running: bool = True) -> None:
        """Sets the tasks associated with a workflow to the appropriate states."""
        self.wait_for_workflow_resume()

        self.requester.send_request(
            app_route=f"/task/{self.workflow_id}/set_resume_state",
            message={"reset_if_running": reset_if_running},
            request_type="post",
        )

    def create_workflow_run(
        self,
        workflow_run_heartbeat_interval: Optional[int] = None,
        heartbeat_report_by_buffer: Optional[float] = None,
    ) -> WorkflowRun:
        """Workflow should at least have signalled for a resume at this point."""
        # create workflow run
        client_wfr = WorkflowRun(
            workflow_id=self.workflow_id,
            requester=self.requester,
            workflow_run_heartbeat_interval=workflow_run_heartbeat_interval,
            heartbeat_report_by_buffer=heartbeat_report_by_buffer,
        )
        client_wfr.bind()

        return client_wfr


class WorkflowRun(object):
    """WorkflowRun enables tracking for multiple runs of a single Workflow.

    A Workflow may be started/paused/ and resumed multiple times. Each start or
    resume represents a new WorkflowRun.

    In order for a Workflow can be deemed to be DONE (successfully), it
    must have 1 or more WorkflowRuns. In the current implementation, a Workflow
    Job may belong to one or more WorkflowRuns, but once the Job reaches a DONE
    state, it will no longer be added to a subsequent WorkflowRun. However,
    this is not enforced via any database constraints.
    """

    def __init__(
        self,
        workflow_id: int,
        requester: Optional[Requester] = None,
        workflow_run_heartbeat_interval: Optional[int] = None,
        heartbeat_report_by_buffer: Optional[float] = None,
    ) -> None:
        """Initialize client WorkflowRun."""
        # set attrs
        self.workflow_id = workflow_id
        self.user = getpass.getuser()

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        # set values from config
        config = JobmonConfig()
        if workflow_run_heartbeat_interval is None:
            heartbeat_interval = config.get_int("heartbeat", "workflow_run_interval")
        else:
            heartbeat_interval = int(workflow_run_heartbeat_interval)
        self.heartbeat_interval = heartbeat_interval
        if heartbeat_report_by_buffer is None:
            report_by_buffer = config.get_float("heartbeat", "report_by_buffer")
        else:
            report_by_buffer = float(heartbeat_report_by_buffer)
        self.heartbeat_report_by_buffer = report_by_buffer

        self._workflow_run_id = None
        self._status: Optional[str] = None

    @property
    def workflow_run_id(self) -> int:
        if not self._workflow_run_id:
            raise WorkflowNotResumable(
                "This workflow run was not yet bound successfully, "
                "cannot access workflow run id attribute."
            )
        return self._workflow_run_id

    @property
    def status(self) -> Optional[str]:
        if not self._status:
            raise WorkflowNotResumable(
                "This workflow run was not bound successfully, "
                "cannot access status attribute."
            )
        return self._status

    def bind(self) -> None:
        """Link this workflow run with the workflow and add all tasks."""
        if self._workflow_run_id:
            return  # WorkflowRun already bound

        next_report_increment = (
            self.heartbeat_interval * self.heartbeat_report_by_buffer
        )

        # bind to database
        app_route = "/workflow_run"
        _, resp = self.requester.send_request(
            app_route=app_route,
            message={
                "workflow_id": self.workflow_id,
                "user": self.user,
                "jobmon_version": __version__,
                "next_report_increment": next_report_increment,
            },
            request_type="post",
        )
        workflow_run_id = resp.get("workflow_run_id")
        if not workflow_run_id:
            raise WorkflowNotResumable(resp.get("err_msg"))
        self._workflow_run_id = workflow_run_id
        self._status = resp.get("status")

    def _update_status(self, status: str) -> None:
        """Update the status of the workflow_run with whatever status is passed."""
        app_route = f"/workflow_run/{self.workflow_run_id}/update_status"
        self.requester.send_request(
            app_route=app_route,
            message={"status": status},
            request_type="put",
        )
        self._status = status

    def __repr__(self) -> str:
        """A representation string for a client WorkflowRun instance."""
        return (
            f"WorkflowRun(workflow_id={self.workflow_id}, "
            f"workflow_run_id={self.workflow_run_id}"
        )



"""The overarching framework to create tasks and dependencies within."""

from __future__ import annotations

import copy
import hashlib
import itertools
import logging
import logging.config
import os
from subprocess import PIPE, Popen, TimeoutExpired
import sys
from types import TracebackType
from typing import Any, Dict, Iterator, List, Optional, Sequence, TYPE_CHECKING, Union
import uuid

import psutil

from jobmon.client.array import Array
from jobmon.client.dag import Dag
from jobmon.client.logging import JobmonLoggerConfig
from jobmon.client.swarm.workflow_run import WorkflowRun as SwarmWorkflowRun
from jobmon.client.task import Task
from jobmon.client.task_resources import TaskResources
from jobmon.client.tool_version import ToolVersion
from jobmon.client.workflow_run import WorkflowRunFactory
from jobmon.core.cluster import Cluster
from jobmon.core.configuration import JobmonConfig
from jobmon.core.constants import (
    MaxConcurrentlyRunning,
    TaskStatus,
    WorkflowRunStatus,
    WorkflowStatus,
)
from jobmon.core.exceptions import (
    ConfigError,
    DistributorStartupTimeout,
    DuplicateNodeArgsError,
    WorkflowAlreadyComplete,
    WorkflowAlreadyExists,
)
from jobmon.core.requester import Requester

if TYPE_CHECKING:
    from jobmon.client.tool import Tool


logger = logging.getLogger(__name__)


class DistributorContext:
    def __init__(self, cluster_name: str, workflow_run_id: int, timeout: int) -> None:
        """Initialization of the DistributorContext."""
        self._cluster_name = cluster_name
        self._workflow_run_id = workflow_run_id
        self._timeout = timeout

    def __enter__(self) -> DistributorContext:
        """Starts the Distributor Process."""
        logger.info("Starting Distributor Process")

        # construct env
        env = os.environ.copy()
        entry_point = self.derive_jobmon_command_from_env()
        if entry_point is not None:
            env["JOBMON__DISTRIBUTOR__WORKER_NODE_ENTRY_POINT"] = f'"{entry_point}"'

        # Start the distributor. Write stderr to a file.
        cmd = [
            sys.executable,
            "-m",  # safest way to find the entrypoint
            "jobmon.distributor.cli",
            "start",
            "--cluster_name",
            self._cluster_name,
            "--workflow_run_id",
            str(self._workflow_run_id),
        ]
        self.process = Popen(
            cmd,
            stderr=PIPE,
            universal_newlines=True,
            env=env,
        )

        # check if stderr contains "ALIVE"
        assert self.process.stderr is not None  # keep mypy happy on optional type
        stderr_val = self.process.stderr.read(5)
        if stderr_val != "ALIVE":
            err = self._shutdown()
            raise DistributorStartupTimeout(
                f"Distributor process did not start, stderr='{err}'"
            )
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[TracebackType],
    ) -> None:
        """Stops the Distributor Process."""
        logger.info("Stopping Distributor Process")
        err = self._shutdown()
        logger.info(f"Got {err} from Distributor Process")

    def alive(self) -> bool:
        self.process.poll()
        return self.process.returncode is None

    def _shutdown(self) -> str:
        self.process.terminate()
        try:
            _, err = self.process.communicate(timeout=self._timeout)
        except TimeoutExpired:
            err = ""

        if "SHUTDOWN" not in err:
            try:
                parent = psutil.Process(self.process.pid)
                for child in parent.children(recursive=True):
                    child.kill()
            except psutil.NoSuchProcess:
                pass
            self.process.kill()
            self.process.wait()

        return err

    @staticmethod
    def derive_jobmon_command_from_env() -> Optional[str]:
        """If a singularity path is provided, use it when running the worker node."""
        singularity_img_path = os.environ.get("IMGPATH", None)
        if singularity_img_path:
            return f"singularity run --app jobmon_command {singularity_img_path}"
        return None


class Workflow(object):
    """(aka Batch, aka Swarm).

    A Workflow is a framework by which a user may define the relationship
    between tasks and define the relationship between multiple runs of the same
    set of tasks. The great benefit of the Workflow is that it's resumable.
    A Workflow can only be re-loaded if two things are shown to be exact
    matches to a previous Workflow:

    1. WorkflowArgs: It is recommended to pass a meaningful unique identifier
        to workflow_args, to ease resuming. However, if the Workflow is a
        one-off project, you may instantiate the Workflow anonymously, without
        WorkflowArgs. Under the hood, the WorkflowArgs will default to a UUID
        which, as it is randomly generated, will be harder to remember and thus
        harder to resume.

        Workflow args must be hashable. For example, CodCorrect or Como version
        might be passed as Args to the Workflow. For now, the assumption is
        WorkflowArgs is a string.

    2. The tasks added to the workflow. A Workflow is built up by
        using Workflow.add_task(). In order to resume a Workflow, all the same
        tasks must be added with the same dependencies between tasks.
    """

    def __init__(
        self,
        tool_version: ToolVersion,
        workflow_args: str = "",
        name: str = "",
        description: str = "",
        workflow_attributes: Optional[Union[List, dict]] = None,
        max_concurrently_running: int = MaxConcurrentlyRunning.MAXCONCURRENTLYRUNNING,
        requester: Optional[Requester] = None,
        chunk_size: int = 500,  # TODO: should be in the config
    ) -> None:
        """Initialization of the client workflow.

        Args:
            tool_version: ToolVersion this workflow is associated
            workflow_args: Unique identifier of a workflow
            name: Name of the workflow
            description: Description of the workflow
            workflow_attributes: Attributes that make this workflow different from other
                workflows that the user wants to record.
            max_concurrently_running: How many running jobs to allow in parallel
            requester: object to communicate with the flask services.
            chunk_size: how many tasks to bind in a single request
            default_max_attempts: the default max attempts of the workflow for each array
        """
        self._tool_version = tool_version
        self.name = name
        self.description = description
        self.max_concurrently_running: int = max_concurrently_running

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        self._dag = Dag(requester)
        # hash to task object mapping. ensure only 1
        self.tasks: Dict[int, Task] = {}
        self.arrays: Dict[str, Array] = {}
        self._chunk_size: int = chunk_size

        if workflow_args:
            self.workflow_args = workflow_args
        else:
            self.workflow_args = str(uuid.uuid4())
            logger.info(
                "Workflow_args defaulting to uuid {}. To resume this "
                "workflow, you must re-instantiate Workflow and pass "
                "this uuid in as the workflow_args. As a uuid is hard "
                "to remember, we recommend you name your workflows and"
                " make workflow_args a meaningful unique identifier. "
                "Then add the same tasks to this workflow".format(self.workflow_args)
            )
        self.workflow_args_hash = int(
            hashlib.sha256(self.workflow_args.encode("utf-8")).hexdigest(), 16
        )

        self.workflow_attributes: Dict[str, Any] = {}
        if workflow_attributes:
            if isinstance(workflow_attributes, List):
                for attr in workflow_attributes:
                    self.workflow_attributes[attr] = None
            elif isinstance(workflow_attributes, dict):
                for attr, val in workflow_attributes.items():
                    self.workflow_attributes[str(attr)] = str(val)
            else:
                raise ValueError(
                    "workflow_attributes must be provided as a list of attributes or a "
                    "dictionary of attributes and their values"
                )

        # Cache for clusters and task resources
        self._clusters: Dict[str, Cluster] = {}
        self._task_resources: Dict[int, TaskResources] = {}
        self.default_cluster_name: str = ""
        self._default_max_attempts: Optional[int] = None
        self.default_compute_resources_set: Dict[str, Dict[str, Any]] = {}
        self.default_resource_scales_set: Dict[str, Dict[str, float]] = {}

        self._fail_after_n_executions = 1_000_000_000
        self.last_workflow_run_id: Optional[int] = None

    @property
    def tool(self) -> Tool:
        """Returns the associated tool to this workflow."""
        return self._tool_version.tool

    @property
    def is_bound(self) -> bool:
        """If the workflow has been bound to the db."""
        if not hasattr(self, "_workflow_id"):
            return False
        else:
            return True

    @property
    def workflow_id(self) -> int:
        """If the workflow is bound then it will have been given an id."""
        if not self.is_bound:
            raise AttributeError(
                "workflow_id cannot be accessed before workflow is bound"
            )
        return self._workflow_id

    @property
    def dag_id(self) -> int:
        """If it has been bound, it will have an associated dag_id."""
        if not self.is_bound:
            raise AttributeError("dag_id cannot be accessed before workflow is bound")
        return self._dag.dag_id

    @property
    def task_hash(self) -> int:
        """Hash of all of the tasks."""
        hash_value = hashlib.sha256()
        tasks = sorted(self.tasks.values())
        if len(tasks) > 0:  # if there are no tasks, we want to skip this
            for task in tasks:
                hash_value.update(str(hash(task)).encode("utf-8"))
        return int(hash_value.hexdigest(), 16)

    @property
    def task_errors(self) -> Dict:
        """Return a dict of error associated with a task."""
        return {
            task.name: task.get_errors()
            for task in self.tasks.values()
            if task.final_status == TaskStatus.ERROR_FATAL
        }

    @property
    def default_max_attempts(self) -> Optional[int]:
        """Return the workflow default max attempts."""
        if self._default_max_attempts is None:
            self._default_max_attempts = self.tool.default_max_attempts
        return self._default_max_attempts

    def add_attributes(self, workflow_attributes: dict) -> None:
        """Users can call either to update values of existing attributes or add new attributes.

        Args:
            workflow_attributes: attributes to be bound to the db that describe
                this workflow.
        """
        app_route = f"/workflow/{self.workflow_id}/workflow_attributes"
        self.requester.send_request(
            app_route=app_route,
            message={"workflow_attributes": workflow_attributes},
            request_type="put",
        )

    def add_task(self, task: Task) -> Task:
        """Add a task to the workflow to be executed.

        Set semantics - add tasks once only, based on hash name.

        Args:
            task: single task to add.
        """
        logger.debug(f"Adding Task {task}")
        if hash(task) in self.tasks.keys():
            raise ValueError(
                f"A task with hash {hash(task)} already exists. "
                f"All tasks in a workflow must have unique "
                f"commands. Your command was: {task.command}"
            )

        try:
            # link array
            self._link_array_and_workflow(task.array)
        except AttributeError:
            # or infer if not already created
            template_name = task.node.task_template_version.task_template.template_name
            try:
                array = self.arrays[template_name]
            except KeyError:
                # create array from the task template version on the node
                array = Array(
                    task_template_version=task.node.task_template_version,
                    task_args=task.task_args,
                    op_args=task.op_args,
                    cluster_name=task.cluster_name,
                    requester=self.requester,
                )
                self._link_array_and_workflow(array)

            # add task to inferred array
            array.add_task(task)
        except ValueError:
            # check if current task array is the same as the one attached to the workflow
            template_name = task.node.task_template_version.task_template.template_name
            if self.arrays[template_name] != task.array:
                raise
        # set array max_attempts
        # task.array.max_attempts = self._default_max_attempts
        # add node to task
        try:
            self._dag.add_node(task.node)
        except DuplicateNodeArgsError:
            raise DuplicateNodeArgsError(
                "All tasks for a given task template in a workflow must have unique node_args."
                f"Found duplicate node args for {task}. task_template_version_id="
                f"{task.node.task_template_version_id}, node_args={task.node.node_args}"
            )

        # add task to workflow
        self.tasks[hash(task)] = task
        task.workflow = self

        logger.debug(f"Task {hash(task)} added")

        return task

    def _link_array_and_workflow(self, array: Array) -> None:
        template_name = array.task_template_version.task_template.template_name
        if template_name in self.arrays.keys():
            raise ValueError(
                f"An array for template_name={template_name} already exists on this workflow."
                f" You can only call TaskTemplate.create_tasks once per task template."
            )
        # add the references
        self.arrays[template_name] = array
        array.workflow = self

    def add_tasks(self, tasks: Sequence[Task]) -> None:
        """Add a list of task to the workflow to be executed."""
        for task in tasks:
            # add the task
            self.add_task(task)

    def set_default_compute_resources_from_yaml(
        self, cluster_name: str, yaml_file: str
    ) -> None:
        """Set default compute resources from a user provided yaml file for workflow level.

        TODO: Implement this method.

        Args:
            cluster_name: name of cluster to set default values for.
            yaml_file: the yaml file that is providing the compute resource values.
        """
        pass

    def set_default_compute_resources_from_dict(
        self, cluster_name: str, dictionary: Dict[str, Any]
    ) -> None:
        """Set default compute resources for a given cluster_name.

        Args:
            cluster_name: name of cluster to set default values for.
            dictionary: dictionary of default compute resources to run tasks
                with. Can be overridden at task template, tool or task level.
        """
        # TODO: Do we need to handle the scenario where no cluster name is specified?
        self.default_compute_resources_set[cluster_name] = dictionary

    def set_default_resource_scales_from_dict(
        self, cluster_name: str, dictionary: Dict[str, float]
    ) -> None:
        """Set default resource scales for a given cluster_name.

        Args:
            cluster_name: name of cluster to set default values for.
            dictionary: dictionary of default resource scales to adjust task
                resources with. Can be overridden at task template or task level.
        """
        # TODO: Do we need to handle the scenario where no cluster name is specified?
        self.default_resource_scales_set[cluster_name] = dictionary

    def set_default_cluster_name(self, cluster_name: str) -> None:
        """Set the default cluster.

        Args:
            cluster_name: name of cluster to set as default.
        """
        self.default_cluster_name = cluster_name

    def set_default_max_attempts(self, value: int) -> None:
        """Set the max attempts.

        Args:
            value: value of max_attempts.
        """
        self._default_max_attempts = value

    def get_tasks_by_node_args(
        self, task_template_name: str, **kwargs: Any
    ) -> List[Task]:
        """Query tasks by node args. Used for setting dependencies."""
        try:
            array = self.arrays[task_template_name]
        except KeyError:
            raise ValueError(
                f"task_template_name={task_template_name} not found on workflow. Known "
                f"template_names are {self.arrays.keys()}."
            )
        tasks = array.get_tasks_by_node_args(**kwargs)
        return tasks

    def set_max_concurrently_running(
        self, task_template_name: str, max_concurrently_running: int
    ) -> None:
        pass

    def run(
        self,
        fail_fast: bool = False,
        seconds_until_timeout: int = 36000,
        resume: bool = False,
        reset_running_jobs: bool = True,
        distributor_startup_timeout: int = 180,
        resume_timeout: int = 300,
        configure_logging: bool = False,
    ) -> Optional[str]:
        """Run the workflow.

        Traverse the dag and submitting new tasks when their tasks have completed successfully.

        Args:
            fail_fast: whether to break out of distributor on first failure.
            seconds_until_timeout: amount of time (in seconds) to wait
                until the whole workflow times out. Submitted jobs will
                continue
            resume: whether the workflow should be resumed or not, if
                it is not set to resume and an identical workflow already
                exists, the workflow will error out
            reset_running_jobs: whether or not to reset running jobs upon resume
            distributor_startup_timeout: amount of time to wait for the distributor process to
                start up
            resume_timeout: seconds to wait for a workflow to become resumable before giving up
            configure_logging: setup jobmon logging. If False, no logging will be configured.
                If True, default logging will be configured.

        Returns:
            str of WorkflowRunStatus
        """
        if configure_logging is True:
            JobmonLoggerConfig.attach_default_handler(
                logger_name="jobmon.client", log_level=logging.INFO
            )

        # bind to database
        logger.info("Adding Workflow metadata to database")
        self.bind()

        config = JobmonConfig()
        try:
            gui_url = config.get("http", "gui_url")
        except ConfigError:
            gui_url = ""

        logger.info(
            f"Workflow ID {self.workflow_id} assigned. Progress can be monitored at "
            f"{gui_url}/#/workflow/{self.workflow_id}/tasks"
        )

        # Check if this workflow is already complete and is runnable
        if self._status == WorkflowStatus.DONE:
            raise WorkflowAlreadyComplete(
                f"Workflow ({self.workflow_id}) is already in done state and cannot be resumed"
            )

        if not self._newly_created and not resume:
            raise WorkflowAlreadyExists(
                "This workflow already exists. If you are trying to resume a workflow, "
                "please set the resume flag. If you are not trying to resume a workflow, make "
                "sure the workflow args are unique or the tasks are unique"
            )
        if self._newly_created and resume:
            logger.warning(
                "The resume flag has been set but no previous workflow_args exist."
                "Note that the workflow will execute as a new workflow."
            )

        # Bind tasks
        logger.info("Adding task metadata to database")
        # Need to wait for resume signal to be sent before resetting tasks, in case of a resume
        factory = WorkflowRunFactory(self.workflow_id)
        if not self._newly_created and resume:
            factory.set_workflow_resume(
                reset_running_jobs=reset_running_jobs, resume_timeout=resume_timeout
            )

        self._bind_tasks(
            reset_if_running=reset_running_jobs, chunk_size=self._chunk_size
        )

        # create workflow_run
        logger.info("Adding WorkflowRun metadata to database")
        wfr = factory.create_workflow_run()
        # Update the workflowrun to BOUND state immediately in this API. All metadata already
        # bound, so the swarm can start immediately.
        wfr._update_status(WorkflowRunStatus.BOUND)
        logger.info(f"WorkflowRun ID {wfr.workflow_run_id} assigned")

        # start distributor
        cluster_name = list(self._clusters.keys())[0]
        with DistributorContext(
            cluster_name, wfr.workflow_run_id, distributor_startup_timeout
        ) as distributor:
            # set up swarm and initial DAG
            swarm = SwarmWorkflowRun(
                workflow_run_id=wfr.workflow_run_id,
                fail_after_n_executions=self._fail_after_n_executions,
                requester=self.requester,
                fail_fast=fail_fast,
                status=wfr.status,
            )
            swarm.from_workflow(self)
            self._num_previously_completed = swarm.num_previously_complete

            try:
                swarm.run(distributor.alive, seconds_until_timeout)
            finally:
                # figure out doneness
                num_new_completed = (
                    len(swarm.done_tasks) - swarm.num_previously_complete
                )
                if swarm.status != WorkflowRunStatus.DONE:
                    logger.info(
                        f"WorkflowRun execution ended, num failed {len(swarm.failed_tasks)}"
                    )
                else:
                    logger.info(
                        f"WorkflowRun execute finished successfully, {num_new_completed} tasks"
                    )

                # update workflow tasks with final status
                for task in self.tasks.values():
                    task.final_status = swarm.tasks[task.task_id].status
                self._num_newly_completed = num_new_completed

        self.last_workflow_run_id = wfr.workflow_run_id

        return swarm.status

    def set_task_template_max_concurrency_limit(
        self, task_template_name: str, limit: int
    ) -> None:
        try:
            array = self.arrays[task_template_name]
        except Exception:
            raise KeyError(
                f"There is no task_template named '{task_template_name}' "
                f"associated with this workflow. Workflow name: {self.name}"
            )
        array.max_concurrently_running = limit

    def validate(self, strict: bool = True, raise_on_error: bool = False) -> None:
        """Confirm that the tasks in this workflow are valid.

        This method will:
        - access the database to confirm the requested resources are valid for
        the specified cluster
        - confirm that the workflow args are valid
        - make sure no task contains up/down stream tasks that are not in the workflow
        """
        # construct task resources
        for task in self.tasks.values():
            # get the cluster for this task
            cluster = self.get_cluster_by_name(task.cluster_name)

            # not dynamic resource request. Construct TaskResources
            if task.compute_resources_callable is None:
                try:
                    queue = cluster.get_queue(task.queue_name)
                except ValueError as e:
                    if raise_on_error:
                        raise e
                    else:
                        logger.info(e)
                        continue

                # validate the constructed resources
                task_resources = TaskResources(
                    requested_resources=task.requested_resources, queue=queue
                )

                is_valid, msg = task_resources.validate_resources(strict)
                if not is_valid:
                    if raise_on_error:
                        raise ValueError(f"Failed validation, reasons: {msg}")
                    else:
                        logger.info(f"Failed validation, reasons: {msg}")

        for array in self.arrays.values():
            try:
                array.validate()
            except ValueError as e:
                if raise_on_error:
                    raise
                else:
                    logger.info(e)
        try:
            cluster_names = list(self._clusters.keys())
            if len(list(self._clusters.keys())) > 1:
                raise RuntimeError(
                    f"Workflow can only use one cluster. Found cluster_names={cluster_names}"
                )
            # check if workflow is valid
            self._dag.validate()
            self._matching_wf_args_diff_hash()
        except Exception as e:
            if raise_on_error:
                raise
            else:
                logger.info(e)

    def bind(self) -> None:
        """Get a workflow_id."""
        if self.is_bound:
            return

        # strict = False means we can coerce. obviously we need to raise at this point
        self.validate(strict=False, raise_on_error=True)

        # bind dag
        self._dag.bind(self._chunk_size)

        # bind workflow
        app_route = "/workflow"
        return_code, response = self.requester.send_request(
            app_route=app_route,
            message={
                "tool_version_id": self._tool_version.id,
                "dag_id": self._dag.dag_id,
                "workflow_args_hash": self.workflow_args_hash,
                "task_hash": self.task_hash,
                "description": self.description,
                "name": self.name,
                "workflow_args": self.workflow_args,
                "max_concurrently_running": self.max_concurrently_running,
                "workflow_attributes": self.workflow_attributes,
            },
            request_type="post",
        )

        self._workflow_id = response["workflow_id"]
        self._status = response["status"]
        self._newly_created = response["newly_created"]

    def _bind_tasks(
        self,
        reset_if_running: bool = True,
        chunk_size: int = 500,
    ) -> None:
        app_route = "/task/bind_tasks_no_args"
        remaining_task_hashes = list(self.tasks.keys())

        while remaining_task_hashes:
            # split off first chunk elements from queue.
            task_hashes_chunk = remaining_task_hashes[:chunk_size]
            remaining_task_hashes = remaining_task_hashes[chunk_size:]

            # If this is the last chunk, mark the created_date field in the
            # database.
            mark_created = len(remaining_task_hashes) == 0

            # send to server in a format of:
            # {<hash>:[workflow_id(0), node_id(1), task_args_hash(2), array_id(3),
            # name(4), command(5), max_attempts(6)], reset_if_running(7), task_args(8),
            # task_attributes(9), resource_scales(10), fallback_queues(11)}
            # flat the data structure so that the server won't depend on the client
            task_metadata: Dict[int, List] = {}
            for task_hash in task_hashes_chunk:
                task = self.tasks[task_hash]

                # get array id
                array = task.array
                if not array.is_bound:
                    array.bind()

                # get task resources id
                self._set_original_task_resources(task)

                serializable_resource_scales = copy.copy(task.resource_scales)
                for resource, scaler in task.resource_scales.items():
                    # We can't serialize a callable, so use the function name instead.
                    if callable(scaler):
                        serializable_resource_scales[resource] = getattr(  # type: ignore
                            scaler, "__name__", "Unknown Callable"  # type: ignore
                        )
                    # We can't serialize an iterator, so take the relevant elements as a
                    # list.
                    elif isinstance(scaler, Iterator):
                        serializable_resource_scales[resource] = list(  # type: ignore
                            itertools.islice(
                                copy.deepcopy(scaler), task.max_attempts - 1
                            )
                        )

                task_metadata[task_hash] = [
                    task.node.node_id,
                    str(task.task_args_hash),
                    task.array.array_id,
                    task.original_task_resources.id,
                    task.name,
                    task.command,
                    task.max_attempts,
                    reset_if_running,
                    serializable_resource_scales,
                    task.fallback_queues,
                ]

            parameters = {
                "workflow_id": self.workflow_id,
                "tasks": task_metadata,
                "mark_created": mark_created,
            }
            return_code, response = self.requester.send_request(
                app_route=app_route,
                message=parameters,
                request_type="put",
            )

            # populate returned values onto task dict
            return_tasks = response["tasks"]
            for k in return_tasks.keys():
                task = self.tasks[int(k)]
                task.task_id = return_tasks[k][0]
                task.initial_status = return_tasks[k][1]

        # Bind task arguments and attributes as well
        self._bind_task_args(chunk_size)
        self._bind_task_attributes(chunk_size)

    def _bind_task_args(self, chunk_size: int = 500) -> None:
        """Bind all task args to the database.

        Loop through our bound task dict in chunks in order to bind new args and arg types
        to the database.
        """
        remaining_task_hashes = list(self.tasks.keys())

        while remaining_task_hashes:
            # split off first chunk elements from queue.
            task_hashes_chunk = remaining_task_hashes[:chunk_size]
            remaining_task_hashes = remaining_task_hashes[chunk_size:]

            task_arg_list = []
            for task_hash in task_hashes_chunk:
                task = self.tasks[task_hash]
                task_args = [
                    (task.task_id, arg_id, value)
                    for arg_id, value in task.mapped_task_args.items()
                ]
                task_arg_list.extend(task_args)

            self.requester.send_request(
                app_route="/task/bind_task_args",
                message={"task_args": task_arg_list},
                request_type="put",
            )

    def _bind_task_attributes(self, chunk_size: int = 500) -> None:
        remaining_task_hashes = list(self.tasks.keys())

        while remaining_task_hashes:
            # split off first chunk elements from queue.
            task_hashes_chunk = remaining_task_hashes[:chunk_size]
            remaining_task_hashes = remaining_task_hashes[chunk_size:]

            attribute_dict = {}
            for task_hash in task_hashes_chunk:
                task = self.tasks[task_hash]
                attribute_dict[task.task_id] = task.task_attributes

            # Send the request
            self.requester.send_request(
                app_route="/task/bind_task_attributes",
                message={"task_attributes": attribute_dict},
                request_type="put",
            )

    def get_errors(
        self, limit: int = 1000
    ) -> Optional[Dict[int, Dict[str, Union[int, List[Dict[str, Union[str, int]]]]]]]:
        """Method to get all errors.

        Return a dictionary with the erring task_id as the key, and
        the Task.get_errors content as the value.
        When limit is specifically set as None from the client, this
        return set will pass back all the erred tasks in the workflow.
        """
        errors = {}

        cnt: int = 0
        for task in self.tasks.values():
            task_id = task.task_id
            task_errors = task.get_errors()
            if task_errors is not None and len(task_errors) > 0:
                errors[task_id] = task_errors
                cnt += 1
                if limit is not None and cnt >= limit - 1:
                    break

        return errors

    def get_cluster_by_name(self, cluster_name: str) -> Cluster:
        """Check if the cluster that the task specified is in the cache.

        If the cluster is not in the cache, create it and add to cache.
        """
        try:
            cluster = self._clusters[cluster_name]
        except KeyError:
            cluster = Cluster(cluster_name=cluster_name, requester=self.requester)
            cluster.bind()
            self._clusters[cluster_name] = cluster
        return cluster

    def _set_original_task_resources(self, task: Task) -> None:
        cluster = self.get_cluster_by_name(task.cluster_name)
        queue = cluster.get_queue(task.queue_name)
        task_resources = TaskResources(
            requested_resources=task.requested_resources, queue=queue
        )

        try:
            task_resources = self._task_resources[hash(task_resources)]
        except KeyError:
            task_resources.bind()
            self._task_resources[hash(task_resources)] = task_resources

        task.original_task_resources = task_resources

    def _matching_wf_args_diff_hash(self) -> None:
        """Check that that an existing workflow does not contain different tasks.

        Check that an existing workflow with the same workflow_args does not have a
        different hash, this would indicate that thgat the workflow contains different tasks.
        """
        rc, response = self.requester.send_request(
            app_route=f"/workflow/{str(self.workflow_args_hash)}",
            message={},
            request_type="get",
        )
        bound_workflow_hashes = response["matching_workflows"]
        for task_hash, tool_version_id, dag_hash in bound_workflow_hashes:
            match = self._tool_version.id == tool_version_id and (
                str(self.task_hash) != task_hash or str(hash(self._dag)) != dag_hash
            )
            if match:
                raise WorkflowAlreadyExists(
                    "The unique workflow_args already belong to a workflow "
                    "that contains different tasks than the workflow you are "
                    "creating, either change your workflow args so that they "
                    "are unique for this set of tasks, or make sure your tasks"
                    " match the workflow you are trying to resume"
                )

    def __hash__(self) -> int:
        """Hash to encompass tool version id, workflow args, tasks and dag."""
        hash_value = hashlib.sha256()
        hash_value.update(str(hash(self._tool_version.id)).encode("utf-8"))
        hash_value.update(str(self.workflow_args_hash).encode("utf-8"))
        hash_value.update(str(self.task_hash).encode("utf-8"))
        hash_value.update(str(hash(self._dag)).encode("utf-8"))
        return int(hash_value.hexdigest(), 16)

    def __repr__(self) -> str:
        """A representation string for a Workflow instance."""
        repr_string = (
            f"Workflow(workflow_args={self.workflow_args}, " f"name={self.name}"
        )
        try:
            repr_string += f", workflow_id={self.workflow_id})"
        except AttributeError:
            # Workflow not yet bound so no ID to add to repr
            repr_string += ")"
        return repr_string



"""The DAG captures the interconnected graph of tasks and their dependencies."""

import hashlib
from http import HTTPStatus as StatusCodes
import logging
from typing import Dict, List, Optional, Set, Tuple, Union

from jobmon.client.node import Node
from jobmon.core.exceptions import (
    CyclicGraphError,
    DuplicateNodeArgsError,
    InvalidResponse,
    NodeDependencyNotExistError,
)
from jobmon.core.requester import Requester

logger = logging.getLogger(__name__)


class Dag(object):
    """The DAG captures the interconnected graph of tasks and their dependencies."""

    def __init__(self, requester: Optional[Requester] = None) -> None:
        """Directed Acyclic Graph.

        The DAG captures the tasks (nodes) as they are
        related to each other in their dependency structure. The Dag is traversed in
        the order of node dependencies so a workflow run is a single instance of
        traversing through a dag. This object stores the nodes and communicates
        with the server with regard to itself.

        Args:
            requester (str): url to communicate with the flask services.
        """
        self.nodes: Set[Node] = set()

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

    @property
    def dag_id(self) -> int:
        """Database unique ID of this DAG."""
        if not hasattr(self, "_dag_id"):
            raise AttributeError("_dag_id cannot be accessed before dag is bound")
        return self._dag_id

    def add_node(self, node: Node) -> None:
        """Add a node to this dag.

        Args:
            node (jobmon.client.node.Node): Node to add to the dag
        """
        # validate node has unique node args within this task template version
        if node in self.nodes:
            raise DuplicateNodeArgsError(
                "A duplicate node was found for task_template_version_id="
                f"{node.task_template_version_id}. Node args were {node.node_args}"
            )
        # wf.add_task should call ClientNode.add_node() + pass the tasks' node
        self.nodes.add(node)

    def bind(self, chunk_size: int = 500) -> int:
        """Retrieve an id for a matching dag from the server.

        If it doesn't exist, first create one, including its edges.
        """
        if len(self.nodes) == 0:
            raise RuntimeError(
                "No nodes were found in the dag. An empty dag " "cannot be bound."
            )

        self._bulk_bind_nodes(chunk_size)

        dag_hash = hash(self)
        _, response = self.requester.send_request(
            app_route="/dag",
            message={"dag_hash": dag_hash},
            request_type="post",
        )
        dag_id = response["dag_id"]

        # no created date means bind edges
        if response["created_date"] is None:
            self._bulk_insert_edges(dag_id)
        self._dag_id = dag_id
        return dag_id

    def validate(self) -> None:
        """Validate the nodes and their dependencies."""
        nodes_in_dag = self.nodes
        for node in nodes_in_dag:
            # Make sure no task contains up/down stream tasks that are not in the workflow
            for n in node.upstream_nodes:
                if n not in nodes_in_dag:
                    raise NodeDependencyNotExistError(
                        f"Upstream node, {hash(n)}, for node, {hash(node)}, "
                        "does not exist in the dag.Check that every task has been added to "
                        "the workflow and is in the correct order."
                    )
            for n in node.downstream_nodes:
                if n not in nodes_in_dag:
                    raise NodeDependencyNotExistError(
                        f"Downstream node, {hash(n)}, for node, {hash(node)}, "
                        "does not exist in the dag.Check that every task has been added to "
                        "the workflow and is in the correct order."
                    )

        dag_map = {node: node.downstream_nodes for node in nodes_in_dag}
        if self._is_cyclic(dag_map):
            raise CyclicGraphError(
                "Cycle detected in the task graph. Please ensure that your task dependencies "
                "flow in only one direction."
            )

    def _is_cyclic(self, dag_map: Dict[Node, Set[Node]]) -> bool:
        """Return true if the nodes are cyclic.

        This method is effectively a depth-first search looking for already-seen nodes,
        implemented using the "stack of iterators" pattern to get around Python's recursion
        limit of 1000.
        """
        visited = set()
        path = [object()]
        path_set = set(path)
        stack = [iter(dag_map)]
        while stack:
            for v in stack[-1]:
                if v in path_set:
                    return True
                elif v not in visited:
                    visited.add(v)
                    path.append(v)
                    path_set.add(v)
                    stack.append(iter(dag_map.get(v, ())))
                    break
            else:
                path_set.remove(path.pop())
                stack.pop()
        return False

    def _bulk_bind_nodes(self, chunk_size: int) -> None:
        def get_chunk(total_nodes: int, chunk_number: int) -> Optional[Tuple[int, int]]:
            # This function is created for unit testing
            if (chunk_number - 1) * chunk_size >= total_nodes:
                return None
            return (
                (chunk_number - 1) * chunk_size,
                min(total_nodes - 1, chunk_number * chunk_size - 1),
            )

        nodes_in_dag = list(self.nodes)
        nodes_received = {}
        total_nodes = len(self.nodes)
        chunk_number = 1
        chunk_boarder = get_chunk(total_nodes, chunk_number)
        while chunk_boarder is not None:
            # do something to bind
            nodes_to_send = []
            for i in range(chunk_boarder[0], chunk_boarder[1] + 1):
                node = nodes_in_dag[i]
                n = {
                    "task_template_version_id": node.task_template_version_id,
                    "node_args_hash": str(node.node_args_hash),
                    "node_args": node.mapped_node_args,
                }
                nodes_to_send.append(n)
            rc, response = self.requester.send_request(
                app_route="/nodes",
                message={"nodes": nodes_to_send},
                request_type="post",
            )
            nodes_received.update(response["nodes"])
            chunk_number += 1
            chunk_boarder = get_chunk(total_nodes, chunk_number)

        for node in nodes_in_dag:
            k = f"{node.task_template_version_id}:{node.node_args_hash}"
            if k in nodes_received.keys():
                node.node_id = int(nodes_received[k])
            else:
                raise InvalidResponse(
                    f"Fail to find node_id in HTTP response for node_args_hash "
                    f"{node.node_args_hash} and task_template_version_id "
                    f"{node.task_template_version_id} HTTP Response:\n {response}"
                )

    def _get_dag_id(self) -> Optional[int]:
        dag_hash = hash(self)
        logger.info(f"Querying for dag with hash: {dag_hash}")
        return_code, response = self.requester.send_request(
            app_route="/dag",
            message={"dag_hash": dag_hash},
            request_type="get",
        )
        if return_code == StatusCodes.OK:
            return response["dag_id"]
        else:
            raise ValueError(
                f"Unexpected status code {return_code} from GET "
                f"request through route /dag/{dag_hash} . "
                f"Expected code 200. Response content: "
                f"{response}"
            )

    def _bulk_insert_edges(self, dag_id: int, chunk_size: int = 500) -> None:
        # compile full list of edges
        all_edges: List[Dict[str, Union[List, int]]] = []
        for node in self.nodes:
            # get the node ids for all upstream and downstream nodes
            upstream_nodes = [
                upstream_node.node_id for upstream_node in node.upstream_nodes
            ]
            downstream_nodes = [
                downstream_node.node_id for downstream_node in node.downstream_nodes
            ]
            all_edges.append(
                {
                    "node_id": node.node_id,
                    "upstream_node_ids": upstream_nodes,
                    "downstream_node_ids": downstream_nodes,
                }
            )
        logger.debug(f"message included in edge post request: {all_edges}")

        while all_edges:
            # split off first chunk elements from queue.
            edge_chunk, all_edges = all_edges[:chunk_size], all_edges[chunk_size:]

            message: Dict[str, Union[List[Dict], bool]] = {"edges_to_add": edge_chunk}
            # more edges to add. don't mark it created
            if all_edges:
                message["mark_created"] = False
            else:
                message["mark_created"] = True

            app_route = f"/dag/{dag_id}/edges"
            self.requester.send_request(
                app_route=app_route, message=message, request_type="post"
            )

    def __hash__(self) -> int:
        """Determined by hashing all sorted node hashes and their downstream."""
        hash_value = hashlib.sha256()
        if len(self.nodes) > 0:  # if the dag is empty, we want to skip this
            for node in sorted(self.nodes):
                hash_value.update(str(hash(node)).encode("utf-8"))
                for downstream_node in sorted(node.downstream_nodes):
                    hash_value.update(str(hash(downstream_node)).encode("utf-8"))
        return int(hash_value.hexdigest(), 16)

    def __repr__(self) -> str:
        """A representation string for a Dag instance."""
        return f"Dag(hash={self.__hash__()}"



from datetime import timedelta
import re
from typing import Tuple

from jobmon.core.exceptions import InvalidMemoryFormat, InvalidMemoryUnit


class TimeUnit:
    """A helper class provides static functions to switch among seconds, minutes, and hours.

    It also can be initialized with combined input of seconds, minutes, and hours,
    and converts them to any single time unit.
    """

    @staticmethod
    def min_to_sec(minutes: float) -> int:
        """Static helper function to turn minutes into seconds. Accept pointers."""
        return int(minutes * 60)

    @staticmethod
    def hour_to_sec(hour: float) -> int:
        """Static helper function to turn minutes into seconds."""
        return int(hour * 60 * 60)

    @staticmethod
    def hour_to_min(hour: float) -> float:
        """Static helper function to turn minutes into seconds."""
        return round(hour * 60.0, 2)

    @staticmethod
    def min_to_hour(min: float) -> float:
        """Static helper function to turn minutes into seconds."""
        return round(min / 60.0, 2)

    @staticmethod
    def sec_to_hour(sec: int) -> float:
        """Static helper function to turn minutes into seconds."""
        return round(sec / 3600.0, 2)

    @staticmethod
    def sec_to_min(sec: int) -> float:
        """Static helper function to turn minutes into seconds."""
        return round(sec / 60.0, 2)

    def __init__(self, sec: int = 0, min: float = 0.0, hour: float = 0.0) -> None:
        """The constructor."""
        self.seconds = sec + TimeUnit.min_to_sec(min) + TimeUnit.hour_to_sec(hour)
        self.minutes = TimeUnit.sec_to_min(sec) + min + TimeUnit.hour_to_min(hour)
        self.hours = TimeUnit.sec_to_hour(sec) + TimeUnit.min_to_hour(min) + hour
        self.readable = str(timedelta(seconds=self.seconds))


class MemUnit:
    """A helper class to convert memory units between B, k, K, m, M, g, G, t, T."""

    base_chart_to_B = {"B": 1, "K": 1024, "M": 1.049e6, "G": 1.074e9, "T": 1.1e12}

    @staticmethod
    def _split_unit(input: str) -> Tuple[int, str]:
        """Split the number and the unit. Return 0 M for invalid; use M when no unit."""
        values = re.findall(r"^\d+", input)
        if len(values) == 0:
            return 0, "M"
        value = values[0]
        units = re.findall(r"\D+$", input)
        if len(units) == 0:
            unit = "M"
        else:
            unit = units[0]
            unit = re.sub("[B|b].*", "B", unit)
            unit = re.sub("[K|k].*", "K", unit)
            unit = re.sub("[M|m].*", "M", unit)
            unit = re.sub("[G|g].*", "G", unit)
            unit = re.sub("[T|t].*", "T", unit)
            if unit not in MemUnit.base_chart_to_B.keys():
                unit = "M"
        return int(value), unit

    @staticmethod
    def _to_B(input: str) -> int:
        value, unit = MemUnit._split_unit(input)
        return int(value * MemUnit.base_chart_to_B[unit])

    @staticmethod
    def convert(input: str, to: str = "M") -> int:
        """Convert memory to different units.

        input: a string in format "1G", "1g", "1Gib", "1gb",etc.
        to: the unit to convert to; takes B, K, M, G, T.
        Lower case unit will be treated as capital.
        Return: an int of the memory value of your specified unit
        """
        input = str(input)
        # match 100, 100m, 100M, 100 Gbetc
        r1 = r"^\d+[bBkKmMgGtT]?[bB]?$"
        # match 100Gib, 100gib, etc
        r2 = r"^\d+[bBkKmMgGtT]i[bB]$"
        if re.search(r1, input) is None and re.search(r2, input) is None:
            raise InvalidMemoryFormat(
                f"Input {input} is invalide. Please use format "
                f"such as 100, 100M, 100Gb, or 100Gib."
            )
        value = MemUnit._to_B(input)
        # apply same input check logic to "to"
        if to.upper() in ("B", "BB", "BIB"):
            to = "B"
        if to.upper() in ("K", "KB", "KIB"):
            to = "K"
        if to.upper() in ("M", "MB", "MIB"):
            to = "M"
        if to.upper() in ("G", "GB", "GIB"):
            to = "G"
        if to.upper() in ("T", "TB", "TIB"):
            to = "T"
        if to not in MemUnit.base_chart_to_B.keys():
            raise InvalidMemoryUnit(
                f"Jobmon can only convert memory unit to "
                f"{MemUnit.base_chart_to_B.keys()}. {to} is invalid."
            )
        return round(value / MemUnit.base_chart_to_B[to])
